###############################################################################
#   EllipticalDiskPotential: Kuijken & Tremaine (1994)'s elliptical disk 
#   potential
###############################################################################
import math as m
from galpy.util import bovy_conversion
from .planarPotential import planarPotential, _APY_LOADED
if _APY_LOADED:
    from astropy import units
_degtorad= m.pi/180.
class EllipticalDiskPotential(planarPotential):
    """Class that implements the Elliptical disk potential of Kuijken & Tremaine (1994) 

    .. math::

        \\Phi(R,\\phi) = \\mathrm{amp}\\,\\phi_0\\,\\left(\\frac{R}{R_1}\\right)^p\\,\\cos\\left(2\\,(\\phi-\\phi_b)\\right)

    This potential can be grown between  :math:`t_{\mathrm{form}}` and  :math:`t_{\mathrm{form}}+T_{\mathrm{steady}}` in a similar way as DehnenBarPotential, but times are given directly in galpy time units

   """
    def __init__(self,amp=1.,phib=25.*_degtorad,
                 p=1.,twophio=0.01,r1=1.,
                 tform=None,tsteady=None,
                 cp=None,sp=None,ro=None,vo=None):
        """
        NAME:

           __init__

        PURPOSE:

           initialize an Elliptical disk potential

           phi(R,phi) = phio (R/Ro)^p cos[2(phi-phib)]

        INPUT:

           amp=  amplitude to be applied to the potential (default:
           1.), see twophio below

           tform= start of growth (to smoothly grow this potential (can be Quantity)

           tsteady= time delay at which the perturbation is fully grown (default: 2.; can be Quantity)

           p= power-law index of the phi(R) = (R/Ro)^p part

           r1= (1.) normalization radius for the amplitude (can be Quantity)

           Either:
           
              a) phib= angle (in rad; default=25 degree; or can be Quantity)

                 twophio= potential perturbation (in terms of 2phio/vo^2 if vo=1 at Ro=1; can be Quantity with units of velocity-squared)
                 
              b) cp, sp= twophio * cos(2phib), twophio * sin(2phib) (can be Quantity with units of velocity-squared)

        OUTPUT:

           (none)

        HISTORY:

           2011-10-19 - Started - Bovy (IAS)

        """
        planarPotential.__init__(self,amp=amp,ro=ro,vo=vo)
        if _APY_LOADED and isinstance(phib,units.Quantity):
            phib= phib.to(units.rad).value
        if _APY_LOADED and isinstance(r1,units.Quantity):
            r1= r1.to(units.kpc).value/self._ro
        if _APY_LOADED and isinstance(tform,units.Quantity):
            tform= tform.to(units.Gyr).value\
                /bovy_conversion.time_in_Gyr(self._vo,self._ro)
        if _APY_LOADED and isinstance(tsteady,units.Quantity):
            tsteady= tsteady.to(units.Gyr).value\
                /bovy_conversion.time_in_Gyr(self._vo,self._ro)
        if _APY_LOADED and isinstance(twophio,units.Quantity):
            twophio= twophio.to(units.km**2/units.s**2).value/self._vo**2.
        if _APY_LOADED and isinstance(cp,units.Quantity):
            cp= cp.to(units.km**2/units.s**2).value/self._vo**2.
        if _APY_LOADED and isinstance(sp,units.Quantity):
            sp= sp.to(units.km**2/units.s**2).value/self._vo**2.
        # Back to old definition
        self._amp/= r1**p
        self.hasC= True
        self.hasC_dxdv= True
        if cp is None or sp is None:
            self._phib= phib
            self._twophio= twophio
        else:
            self._twophio= m.sqrt(cp*cp+sp*sp)
            self._phib= m.atan2(sp,cp)/2.
        self._p= p
        if not tform is None:
            self._tform= tform
        else:
            self._tform= None
        if not tsteady is None:
            self._tsteady= self._tform+tsteady
        else:
            if self._tform is None: self._tsteady= None
            else: self._tsteady= self._tform+2.

    def _evaluate(self,R,phi=0.,t=0.):
        """
        NAME:
           _evaluate
        PURPOSE:
           evaluate the potential at R,phi,t
        INPUT:
           R - Galactocentric cylindrical radius
           phi - azimuth
           t - time
        OUTPUT:
           Phi(R,phi,t)
        HISTORY:
           2011-10-19 - Started - Bovy (IAS)
        """
        #Calculate relevant time
        if not self._tform is None:
            if t < self._tform:
                smooth= 0.
            elif t < self._tsteady:
                deltat= t-self._tform
                xi= 2.*deltat/(self._tsteady-self._tform)-1.
                smooth= (3./16.*xi**5.-5./8*xi**3.+15./16.*xi+.5)
            else: #fully on
                smooth= 1.
        else:
            smooth= 1.
        return smooth*self._twophio/2.*R**self._p\
            *m.cos(2.*(phi-self._phib))
        
    def _Rforce(self,R,phi=0.,t=0.):
        """
        NAME:
           _Rforce
        PURPOSE:
           evaluate the radial force for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           phi - azimuth
           t - time
        OUTPUT:
           the radial force
        HISTORY:
           2011-10-19 - Written - Bovy (IAS)
        """
        #Calculate relevant time
        if not self._tform is None:
            if t < self._tform:
                smooth= 0.
            elif t < self._tsteady:
                deltat= t-self._tform
                xi= 2.*deltat/(self._tsteady-self._tform)-1.
                smooth= (3./16.*xi**5.-5./8*xi**3.+15./16.*xi+.5)
            else: #fully on
                smooth= 1.
        else:
            smooth= 1.
        return -smooth*self._p*self._twophio/2.*R**(self._p-1.)\
            *m.cos(2.*(phi-self._phib))
        
    def _phiforce(self,R,phi=0.,t=0.):
        """
        NAME:
           _phiforce
        PURPOSE:
           evaluate the azimuthal force for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           phi - azimuth
           t - time
        OUTPUT:
           the azimuthal force
        HISTORY:
           2011-10-19 - Written - Bovy (IAS)
        """
        #Calculate relevant time
        if not self._tform is None:
            if t < self._tform:
                smooth= 0.
            elif t < self._tsteady:
                deltat= t-self._tform
                xi= 2.*deltat/(self._tsteady-self._tform)-1.
                smooth= (3./16.*xi**5.-5./8*xi**3.+15./16.*xi+.5)
            else: #fully on
                smooth= 1.
        else:
            smooth= 1.
        return smooth*self._twophio*R**self._p*m.sin(2.*(phi-self._phib))

    def _R2deriv(self,R,phi=0.,t=0.):
        #Calculate relevant time
        if not self._tform is None:
            if t < self._tform:
                smooth= 0.
            elif t < self._tsteady:
                deltat= t-self._tform
                xi= 2.*deltat/(self._tsteady-self._tform)-1.
                smooth= (3./16.*xi**5.-5./8*xi**3.+15./16.*xi+.5)
            else: #fully on
                smooth= 1.
        else:
            smooth= 1.
        return smooth*self._p*(self._p-1.)/2.*self._twophio*R**(self._p-2.)\
            *m.cos(2.*(phi-self._phib))
        
    def _phi2deriv(self,R,phi=0.,t=0.):
        #Calculate relevant time
        if not self._tform is None:
            if t < self._tform:
                smooth= 0.
            elif t < self._tsteady:
                deltat= t-self._tform
                xi= 2.*deltat/(self._tsteady-self._tform)-1.
                smooth= (3./16.*xi**5.-5./8*xi**3.+15./16.*xi+.5)
            else: #perturbation is fully on
                smooth= 1.
        else:
            smooth= 1.
        return -2.*smooth*self._twophio*R**self._p*m.cos(2.*(phi-self._phib))

    def _Rphideriv(self,R,phi=0.,t=0.):
        #Calculate relevant time
        if not self._tform is None:
            if t < self._tform:
                smooth= 0.
            elif t < self._tsteady:
                deltat= t-self._tform
                xi= 2.*deltat/(self._tsteady-self._tform)-1.
                smooth= (3./16.*xi**5.-5./8*xi**3.+15./16.*xi+.5)
            else: #perturbation is fully on
                smooth= 1.
        else:
            smooth= 1.
        return -smooth*self._p*self._twophio*R**(self._p-1.)*m.sin(2.*(phi-self._phib))

    def tform(self): #pragma: no cover
        """
        NAME:

           tform

        PURPOSE:

           return formation time of the perturbation

        INPUT:

           (none)

        OUTPUT:

           tform in normalized units

        HISTORY:

           2011-10-19 - Written - Bovy (IAS)

        """
        return self._tform



# Dehnen & Binney (1998) potentials: models 1 through 4
import numpy
from ..potential import TwoPowerTriaxialPotential
from ..potential import DiskSCFPotential
from ..potential import SCFPotential
from ..potential import scf_compute_coeffs_axi
from ..potential import mwpot_helpers
from ..util import bovy_conversion
# Unit normalization, all models have R0 = 8 kpc
ro= 8.0
# Parameters in common between all models
Rm_ISM= 4./ro
zd_ISM= 0.040/ro
zd_thin= 0.180/ro
zd_thick= 1./ro
Rd_ISM_over_Rd= 2.
Rd_thin_over_Rd= 1.
Rd_thick_over_Rd= 1.
SigmaR0_ISM_over_SigmaR0= 0.25
SigmaR0_thin_over_SigmaR0= 0.70
SigmaR0_thick_over_SigmaR0= 0.05
alpha_bulge= 1.8
q_bulge= 0.6
r0_bulge= 1./ro
rc_bulge= 1.9/ro
q_halo= 0.8

def define_dehnenbinney98_models(model=1):
    # Suppress the numpy floating-point warnings that this code generates...
    old_error_settings= numpy.seterr(all='ignore')
    if model == 1:
        vo= 222.
        Rd_star= 0.25
        Sigma0= 1905.
        rho0_bulge= 0.4271
        rho0_halo= 0.7110
        alpha_halo= -2.
        beta_halo= 2.959
        r0_halo= 3.83/ro
    elif model == 2:
        vo= 217.
        Rd_star= 0.30
        Sigma0= 1208.
        rho0_bulge= 0.7561
        rho0_halo= 1.263
        alpha_halo= -2.
        beta_halo= 2.207
        r0_halo= 1.09/ro
    elif model == 3:
        vo= 217.
        Rd_star= 0.35
        Sigma0= 778.4
        rho0_bulge= 0.3
        rho0_halo= 0.1179
        alpha_halo= 1.8
        beta_halo= 2.002
        r0_halo= 2.29/ro
    elif model == 4:
        vo= 220.
        Rd_star= 0.40
        Sigma0= 536.
        rho0_bulge= 0.3
        rho0_halo= 0.2659
        alpha_halo= 1.629
        beta_halo= 2.167
        r0_halo= 1.899/ro
    sigo= bovy_conversion.surfdens_in_msolpc2(vo=vo,ro=ro)
    rhoo= bovy_conversion.dens_in_msolpc3(vo=vo,ro=ro)
    Sigma0/= sigo
    rho0_bulge/= rhoo
    rho0_halo/= rhoo
    Rd_ISM= Rd_star*Rd_ISM_over_Rd
    Rd_thin= Rd_star*Rd_thin_over_Rd
    Rd_thick= Rd_star*Rd_thick_over_Rd
    SigmaR0= Sigma0/(SigmaR0_ISM_over_SigmaR0\
                         /(2.*zd_ISM)
         /mwpot_helpers.expexp_dens_with_hole(1.,0.,Rd_ISM,Rm_ISM,zd_ISM,1.)
                 +SigmaR0_thin_over_SigmaR0\
         *mwpot_helpers.expexp_dens(0.,0.,Rd_thin,zd_thin,1.)\
         /mwpot_helpers.expexp_dens(1.,0.,Rd_thin,zd_thin,1.)
                 +SigmaR0_thick_over_SigmaR0\
         *mwpot_helpers.expexp_dens(0.,0.,Rd_thick,zd_thick,1.)\
         /mwpot_helpers.expexp_dens(1.,0.,Rd_thick,zd_thick,1.))
    Sigma0_ISM= SigmaR0_ISM_over_SigmaR0*SigmaR0/(2.*zd_ISM)\
        /mwpot_helpers.expexp_dens_with_hole(1.,0.,Rd_ISM,Rm_ISM,zd_ISM,1.)
    Sigma0_thin= SigmaR0_thin_over_SigmaR0*SigmaR0\
        *mwpot_helpers.expexp_dens(0.,0.,Rd_thin,zd_thin,1.)\
        /mwpot_helpers.expexp_dens(1.,0.,Rd_thin,zd_thin,1.)
    Sigma0_thick= SigmaR0_thick_over_SigmaR0*SigmaR0\
        *mwpot_helpers.expexp_dens(0.,0.,Rd_thick,zd_thick,1.)\
        /mwpot_helpers.expexp_dens(1.,0.,Rd_thick,zd_thick,1.)
    # now define gas and disk functions for DiskSCF
    gas_dens= lambda R,z: mwpot_helpers.expexp_dens_with_hole(R,z,Rd_ISM,
                                                              Rm_ISM,zd_ISM,
                                                              Sigma0_ISM)
    disk_dens= lambda R,z: mwpot_helpers.expexp_dens(R,z,Rd_thin,
                                                     zd_thin,Sigma0_thin)\
                      +mwpot_helpers.expexp_dens(R,z,Rd_thick,
                                                 zd_thick,Sigma0_thick)
    bulge_dens= lambda R,z: mwpot_helpers.pow_dens_with_cut(R,z,alpha_bulge,
                                                            r0_bulge,rc_bulge,
                                                            rho0_bulge,q_bulge)
    #dicts used in DiskSCFPotential 
    sigmadict = [{'type':'expwhole','h':Rd_ISM,'amp':Sigma0_ISM,
                  'Rhole':Rm_ISM},
                 {'type':'exp','h':Rd_thin,'amp':Sigma0_thin},
                 {'type':'exp','h':Rd_thick,'amp':Sigma0_thick}]
    hzdict = [{'type':'exp', 'h':zd_ISM},
              {'type':'exp', 'h':zd_thin},
              {'type':'exp', 'h':zd_thick}]
    # Now put together the potential
    DB98_bulge= SCFPotential(\
        Acos=scf_compute_coeffs_axi(bulge_dens,30,10,a=0.025)[0],
        a=0.025,ro=ro,vo=vo)
    DB98_disk= DiskSCFPotential(\
        dens=lambda R,z: gas_dens(R,z)+disk_dens(R,z),
        Sigma=sigmadict,hz=hzdict,a=2.5,N=30,L=30,ro=ro,vo=vo)
    DB98_halo= TwoPowerTriaxialPotential(amp=rho0_halo*(4*numpy.pi*r0_halo**3),
                                         alpha=alpha_halo,beta=beta_halo,
                                         a=r0_halo,ro=ro,vo=vo,c=q_halo)
    # Go back to old floating-point warnings settings
    numpy.seterr(**old_error_settings)
    return DB98_bulge+DB98_disk+DB98_halo




###############################################################################
#   DehnenBarPotential: Dehnen (2000)'s bar potential
###############################################################################
import numpy
from galpy.util import bovy_conversion
from .Potential import Potential, _APY_LOADED
if _APY_LOADED:
    from astropy import units
_degtorad= numpy.pi/180.
class DehnenBarPotential(Potential):
    """Class that implements the Dehnen bar potential (`Dehnen 2000 <http://adsabs.harvard.edu/abs/2000AJ....119..800D>`__), generalized to 3D following `Monari et al. (2016) <http://adsabs.harvard.edu/abs/2016MNRAS.461.3835M>`__

    .. math::

        \\Phi(R,z,\\phi) = A_b(t)\\,\\cos\\left(2\\,(\\phi-\\Omega_b\\,t)\\right))\\,\\left(\\frac{R}{r}\\right)^2\\,\\times \\begin{cases}
        -(R_b/r)^3\\,, & \\text{for}\\ r \\geq R_b\\\\
        (r/R_b)^3-2\\,, & \\text{for}\\ r\\leq R_b.
        \\end{cases}

    where :math:`r^2 = R^2+z^2` is the spherical radius and

    .. math::

        A_b(t) = A_f\\,\\left(\\frac{3}{16}\\xi^5-\\frac{5}{8}\\xi^3+\\frac{15}{16}\\xi+\\frac{1}{2}\\right)\\,, \\xi = 2\\frac{t/T_b-t_\\mathrm{form}}{T_\mathrm{steady}}-1\\,,\ \mathrm{if}\ t_\\mathrm{form} \\leq \\frac{t}{T_b} \\leq t_\\mathrm{form}+T_\\mathrm{steady}

    and 

    .. math::

        A_b(t) = \\begin{cases}
        0\\,, & \\frac{t}{T_b} < t_\mathrm{form}\\\\
        A_f\\,, & \\frac{t}{T_b} > t_\mathrm{form}+T_\mathrm{steady}
        \\end{cases}

    where

    .. math::

       T_b = \\frac{2\pi}{\\Omega_b}

    is the bar period and the strength can also be specified using :math:`\\alpha`

    .. math::

       \\alpha = 3\\,\\frac{A_f}{v_0^2}\\,\\left(\\frac{R_b}{r_0}\\right)^3\,.

    """
    normalize= property() # turn off normalize
    def __init__(self,amp=1.,omegab=None,rb=None,chi=0.8,
                 rolr=0.9,barphi=25.*_degtorad,
                 tform=-4.,tsteady=None,beta=0.,
                 alpha=0.01,Af=None,ro=None,vo=None):
        """
        NAME:

           __init__

        PURPOSE:

           initialize a Dehnen bar potential

        INPUT:

           amp - amplitude to be applied to the potential (default:
           1., see alpha or Ab below)

           barphi - angle between sun-GC line and the bar's major axis
           (in rad; default=25 degree; or can be Quantity))

           tform - start of bar growth / bar period (default: -4)

           tsteady - time from tform at which the bar is fully grown / bar period (default: -tform/2, st the perturbation is fully grown at tform/2)

           Either provide:

              a) rolr - radius of the Outer Lindblad Resonance for a
                 circular orbit (can be Quantity)
              
                 chi - fraction R_bar / R_CR (corotation radius of bar)

                 alpha - relative bar strength (default: 0.01)

                 beta - power law index of rotation curve (to
                 calculate OLR, etc.)
               
              b) omegab - rotation speed of the bar (can be Quantity)
              
                 rb - bar radius (can be Quantity)
                 
                 Af - bar strength (can be Quantity)
              
        OUTPUT:

           (none)

        HISTORY:

           2010-11-24 - Started - Bovy (NYU)

           2017-06-23 - Converted to 3D following Monari et al. (2016) - Bovy (UofT/CCA)

        """
        Potential.__init__(self,amp=amp,ro=ro,vo=vo)
        if _APY_LOADED and isinstance(barphi,units.Quantity):
            barphi= barphi.to(units.rad).value
        if _APY_LOADED and isinstance(rolr,units.Quantity):
            rolr= rolr.to(units.kpc).value/self._ro
        if _APY_LOADED and isinstance(rb,units.Quantity):
            rb= rb.to(units.kpc).value/self._ro
        if _APY_LOADED and isinstance(omegab,units.Quantity):
            omegab= omegab.to(units.km/units.s/units.kpc).value\
                /bovy_conversion.freq_in_kmskpc(self._vo,self._ro)
        if _APY_LOADED and isinstance(Af,units.Quantity):
            Af= Af.to(units.km**2/units.s**2).value/self._vo**2.
        self.hasC= True
        self.hasC_dxdv= True
        self.isNonAxi= True
        self._barphi= barphi
        if omegab is None:
            self._rolr= rolr
            self._chi= chi
            self._beta= beta
            #Calculate omegab and rb
            self._omegab= 1./((self._rolr**(1.-self._beta))/(1.+numpy.sqrt((1.+self._beta)/2.)))
            self._rb= self._chi*self._omegab**(1./(self._beta-1.))
            self._alpha= alpha
            self._af= self._alpha/3./self._rb**3.
        else:
            self._omegab= omegab
            self._rb= rb
            self._af= Af
        self._tb= 2.*numpy.pi/self._omegab
        self._tform= tform*self._tb
        if tsteady is None:
            self._tsteady= self._tform/2.
        else:
            self._tsteady= self._tform+tsteady*self._tb

    def _smooth(self,t):
        if isinstance(t,numpy.ndarray):
            smooth=numpy.ones(len(t))
            indx=(t < self._tform)
            smooth[indx]=0.

            indx=(t < self._tsteady) * (t >= self._tform)
            deltat=t[indx]-self._tform
            xi= 2.*deltat/(self._tsteady-self._tform)-1.
            smooth[indx]= (3./16.*xi**5.-5./8*xi**3.+15./16.*xi+.5)
        else:
            if t < self._tform:
                smooth= 0.
            elif t < self._tsteady:
                deltat= t-self._tform
                xi= 2.*deltat/(self._tsteady-self._tform)-1.
                smooth= (3./16.*xi**5.-5./8*xi**3.+15./16.*xi+.5)
            else: #bar is fully on
                smooth= 1.

        return smooth

    def _evaluate(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _evaluate
        PURPOSE:
           evaluate the potential at R,phi,t
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           Phi(R,z,phi,t)
        HISTORY:
           2010-11-24 - Started - Bovy (NYU)
        """
        #Calculate relevant time
        smooth=self._smooth(t)
        r2= R**2.+z**2.
        r= numpy.sqrt(r2)
        if isinstance(r,numpy.ndarray):
            if not isinstance(R,numpy.ndarray):
                R=numpy.repeat(R,len(r))
            out=numpy.empty(len(r))
            indx= r <= self._rb  
            out[indx]= ((r[indx]/self._rb)**3.-2.)*R[indx]**2./r2[indx]
            indx=numpy.invert(indx)
            out[indx]= -(self._rb/r[indx])**3.*R[indx]**2./r2[indx] 

            out*=self._af*smooth*numpy.cos(2.*(phi-self._omegab*t-self._barphi))
            return out        
        else:
            if r <= self._rb:
                return self._af*smooth*numpy.cos(2.*(phi-self._omegab*t-self._barphi))\
                    *((r/self._rb)**3.-2.)*R**2./r2
            else:
                return -self._af*smooth*numpy.cos(2.*(phi-self._omegab*t-
                                                      self._barphi))\
                                                      *(self._rb/r)**3.\
                                                      *R**2./r2

    def _Rforce(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _Rforce
        PURPOSE:
           evaluate the radial force for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           the radial force
        HISTORY:
           2010-11-24 - Written - Bovy (NYU)
        """
        #Calculate relevant time
        smooth=self._smooth(t)
        r= numpy.sqrt(R**2.+z**2.)
        if isinstance(r,numpy.ndarray):
            if not isinstance(R,numpy.ndarray):
                R=numpy.repeat(R,len(r))
            if not isinstance(z,numpy.ndarray):
                z=numpy.repeat(z,len(r))
            out=numpy.empty(len(r))
            indx= r <= self._rb
            out[indx]= -((r[indx]/self._rb)**3.*R[indx]*(3.*R[indx]**2.+2.*z[indx]**2.)-4.*R[indx]*z[indx]**2.)/r[indx]**4.
            indx= numpy.invert(indx)
            out[indx]= -(self._rb/r[indx])**3.*R[indx]/r[indx]**4.*(3.*R[indx]**2.-2.*z[indx]**2.)

            out*=self._af*smooth*numpy.cos(2.*(phi-self._omegab*t-self._barphi))
            return out
        else:
            if r <= self._rb:
                return -self._af*smooth*numpy.cos(2.*(phi-self._omegab*t
                                                      -self._barphi))\
                        *((r/self._rb)**3.*R*(3.*R**2.+2.*z**2.)-4.*R*z**2.)/r**4.
            else:
                return -self._af*smooth*numpy.cos(2.*(phi-self._omegab*t-
                                                      self._barphi))\
                        *(self._rb/r)**3.*R/r**4.*(3.*R**2.-2.*z**2.)

    def _phiforce(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _phiforce
        PURPOSE:
           evaluate the azimuthal force for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           the azimuthal force
        HISTORY:
           2010-11-24 - Written - Bovy (NYU)
        """
        #Calculate relevant time
        smooth=self._smooth(t)
        r2= R**2.+z**2.
        r= numpy.sqrt(r2)
        if isinstance(r,numpy.ndarray):
            if not isinstance(R,numpy.ndarray):
                R=numpy.repeat(R,len(r))
            if not isinstance(z,numpy.ndarray):
                z=numpy.repeat(z,len(r))
            out=numpy.empty(len(r))
            indx= r <= self._rb  
            out[indx]= ((r[indx]/self._rb)**3.-2.)*R[indx]**2./r2[indx]
            indx=numpy.invert(indx)
            out[indx]= -(self._rb/r[indx])**3.*R[indx]**2./r2[indx]
            out*=2.*self._af*smooth*numpy.sin(2.*(phi-self._omegab*t-self._barphi))
            return out
        else:
            if r <= self._rb:
                return 2.*self._af*smooth*numpy.sin(2.*(phi-self._omegab*t-
                                                        self._barphi))\
                                                    *((r/self._rb)**3.-2.)*R**2./r2
            else:
                return -2.*self._af*smooth*numpy.sin(2.*(phi-self._omegab*t-
                                                         self._barphi))\
                                                         *(self._rb/r)**3.*R**2./r2

    def _zforce(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _zforce
        PURPOSE:
           evaluate the vertical  force for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           the vertical force
        HISTORY:
           2017-06-23 - Written - Bovy (NYU)
        """
        #Calculate relevant time
        smooth=self._smooth(t)
        r= numpy.sqrt(R**2.+z**2.)
        if isinstance(r,numpy.ndarray):
            if not isinstance(R,numpy.ndarray):
                R=numpy.repeat(R,len(r))
            if not isinstance(z,numpy.ndarray):
                z=numpy.repeat(z,len(r))
            out=numpy.empty(len(r))
            indx= r <= self._rb  
            out[indx]= -((r[indx]/self._rb)**3.+4.)*R[indx]**2.*z[indx]/r[indx]**4.
            indx=numpy.invert(indx)
            out[indx]= -5.*(self._rb/r[indx])**3.*R[indx]**2.*z[indx]/r[indx]**4.

            out*=self._af*smooth*numpy.cos(2.*(phi-self._omegab*t-self._barphi))
            return out
        else:
            if r <= self._rb:
                return -self._af*smooth*numpy.cos(2.*(phi-self._omegab*t
                                                      -self._barphi))\
                       *((r/self._rb)**3.+4.)*R**2.*z/r**4.
            else:
                return -5.*self._af*smooth*numpy.cos(2.*(phi-self._omegab*t-
                                                      self._barphi))\
                                *(self._rb/r)**3.*R**2.*z/r**4.

    def _R2deriv(self,R,z,phi=0.,t=0.):
        #Calculate relevant time
        smooth=self._smooth(t)
        r= numpy.sqrt(R**2.+z**2.)
        if isinstance(r,numpy.ndarray):
            if not isinstance(R,numpy.ndarray):
                R=numpy.repeat(R,len(r))
            if not isinstance(z,numpy.ndarray):
                z=numpy.repeat(z,len(r))
            out=numpy.empty(len(r))
            indx= r <= self._rb  
            out[indx]= ((r[indx]/self._rb)**3.*((9.*R[indx]**2.+2.*z[indx]**2.)/r[indx]**4.
                                        -R[indx]**2./r[indx]**6.*(3.*R[indx]**2.+2.*z[indx]**2.))\
                          +4.*z[indx]**2./r[indx]**6.*(4.*R[indx]**2.-r[indx]**2.))
            indx=numpy.invert(indx)
            out[indx]= (self._rb/r[indx])**3./r[indx]**6.*((r[indx]**2.-7.*R[indx]**2.)*(3.*R[indx]**2.-2.*z[indx]**2.)\
                                                 +6.*R[indx]**2.*r[indx]**2.)

            out*=self._af*smooth*numpy.cos(2.*(phi-self._omegab*t-self._barphi))
            return out
        else:
            if r <= self._rb:
                return self._af*smooth*numpy.cos(2.*(phi-self._omegab*t
                                                     -self._barphi))\
                    *((r/self._rb)**3.*((9.*R**2.+2.*z**2.)/r**4.
                                        -R**2./r**6.*(3.*R**2.+2.*z**2.))\
                          +4.*z**2./r**6.*(4.*R**2.-r**2.))
            else:
                return self._af*smooth*numpy.cos(2.*(phi-self._omegab*t-
                                                     self._barphi))\
                    *(self._rb/r)**3./r**6.*((r**2.-7.*R**2.)*(3.*R**2.-2.*z**2.)\
                                                 +6.*R**2.*r**2.)

    def _phi2deriv(self,R,z,phi=0.,t=0.):
        #Calculate relevant time
        smooth=self._smooth(t)
        r= numpy.sqrt(R**2.+z**2.)
        if isinstance(r,numpy.ndarray):
            if not isinstance(R,numpy.ndarray):
                R=numpy.repeat(R,len(r))
            if not isinstance(z,numpy.ndarray):
                z=numpy.repeat(z,len(r))
            out=numpy.empty(len(r))
            indx= r <= self._rb  
            out[indx]= -((r[indx]/self._rb)**3.-2.)*R[indx]**2./r[indx]**2.
            indx=numpy.invert(indx)
            out[indx]= (self._rb/r[indx])**3.*R[indx]**2./r[indx]**2.

            out*=4.*self._af*smooth*numpy.cos(2.*(phi-self._omegab*t-self._barphi))
            return out
        else:
            if r <= self._rb:
                return -4.*self._af*smooth*numpy.cos(2.*(phi-self._omegab*t-
                                                         self._barphi))\
                                                *((r/self._rb)**3.-2.)*R**2./r**2.
            else:
                return 4.*self._af*smooth*numpy.cos(2.*(phi-self._omegab*t-
                                                        self._barphi))\
                                                     *(self._rb/r)**3.*R**2./r**2.

    def _Rphideriv(self,R,z,phi=0.,t=0.):
        #Calculate relevant time
        smooth=self._smooth(t)
        r= numpy.sqrt(R**2.+z**2.)
        if isinstance(r,numpy.ndarray):
            if not isinstance(R,numpy.ndarray):
                R=numpy.repeat(R,len(r))
            if not isinstance(z,numpy.ndarray):
                z=numpy.repeat(z,len(r))
            out=numpy.empty(len(r))
            indx= r <= self._rb
            out[indx]= ((r[indx]/self._rb)**3.*R[indx]*(3.*R[indx]**2.+2.*z[indx]**2.)-4.*R[indx]*z[indx]**2.)/r[indx]**4.
            indx=numpy.invert(indx)
            out[indx]= (self._rb/r[indx])**3.*R[indx]/r[indx]**4.*(3.*R[indx]**2.-2.*z[indx]**2.)

            out*=-2.*self._af*smooth*numpy.sin(2.*(phi-self._omegab*t-self._barphi))
            return out
        else:
            if r <= self._rb:
                return -2.*self._af*smooth*numpy.sin(2.*(phi-self._omegab*t
                                                      -self._barphi))\
                        *((r/self._rb)**3.*R*(3.*R**2.+2.*z**2.)-4.*R*z**2.)/r**4.
            else:
                return -2.*self._af*smooth*numpy.sin(2.*(phi-self._omegab*t-
                                                      self._barphi))\
                        *(self._rb/r)**3.*R/r**4.*(3.*R**2.-2.*z**2.)
             
    def _z2deriv(self,R,z,phi=0.,t=0.):
        #Calculate relevant time
        smooth=self._smooth(t)
        r= numpy.sqrt(R**2.+z**2.)
        if isinstance(r,numpy.ndarray):
            if not isinstance(R,numpy.ndarray):
                R=numpy.repeat(R,len(r))
            if not isinstance(z,numpy.ndarray):
                z=numpy.repeat(z,len(r))
            out=numpy.empty(len(r))
            indx= r <= self._rb
            out[indx]= R[indx]**2./r[indx]**6.*((r[indx]/self._rb)**3.*(r[indx]**2.-z[indx]**2.)
                                     +4.*(r[indx]**2.-4.*z[indx]**2.))
            indx=numpy.invert(indx)
            out[indx]=5.*(self._rb/r[indx])**3.*R[indx]**2./r[indx]**6.*(r[indx]**2.-7.*z[indx]**2.)

            out*=self._af*smooth*numpy.cos(2.*(phi-self._omegab*t-self._barphi))
            return out
        else:
            if r <= self._rb:
                return self._af*smooth*numpy.cos(2.*(phi-self._omegab*t
                                                      -self._barphi))\
                       *R**2./r**6.*((r/self._rb)**3.*(r**2.-z**2.)
                                     +4.*(r**2.-4.*z**2.))
            else:
                return 5.*self._af*smooth*numpy.cos(2.*(phi-self._omegab*t-
                                                      self._barphi))\
                        *(self._rb/r)**3.*R**2./r**6.*(r**2.-7.*z**2.)

    def _Rzderiv(self,R,z,phi=0.,t=0.):
        #Calculate relevant time
        smooth=self._smooth(t)
        r= numpy.sqrt(R**2.+z**2.)
        if isinstance(r,numpy.ndarray):
            if not isinstance(R,numpy.ndarray):
                R=numpy.repeat(R,len(r))
            if not isinstance(z,numpy.ndarray):
                z=numpy.repeat(z,len(r))
            out=numpy.empty(len(r))
            indx= r <= self._rb
            out[indx]= R[indx]*z[indx]/r[indx]**6.*((r[indx]/self._rb)**3.*(2.*r[indx]**2.-R[indx]**2.)
                                     +8.*(r[indx]**2.-2.*R[indx]**2.))
            indx=numpy.invert(indx)
            out[indx]= 5.*(self._rb/r[indx])**3.*R[indx]*z[indx]/r[indx]**6.*(2.*r[indx]**2.-7.*R[indx]**2.)

            out*=self._af*smooth*numpy.cos(2.*(phi-self._omegab*t-self._barphi))
            return out
        else:
            if r <= self._rb:
                return self._af*smooth*numpy.cos(2.*(phi-self._omegab*t
                                                      -self._barphi))\
                       *R*z/r**6.*((r/self._rb)**3.*(2.*r**2.-R**2.)
                                     +8.*(r**2.-2.*R**2.))
            else:
                return 5.*self._af*smooth*numpy.cos(2.*(phi-self._omegab*t-
                                                      self._barphi))\
                        *(self._rb/r)**3.*R*z/r**6.*(2.*r**2.-7.*R**2.)

    def tform(self): #pragma: no cover
        """
        NAME:

           tform

        PURPOSE:

           return formation time of the bar

        INPUT:

           (none)

        OUTPUT:

           tform in normalized units

        HISTORY:

           2011-03-08 - Written - Bovy (NYU)

        """
        return self._tform

    def OmegaP(self):
        """
        NAME:


           OmegaP

        PURPOSE:

           return the pattern speed

        INPUT:

           (none)

        OUTPUT:

           pattern speed

        HISTORY:

           2011-10-10 - Written - Bovy (IAS)

        """
        return self._omegab



###############################################################################
#   actionAngle: a Python module to calculate  actions, angles, and frequencies
#
#      class: actionAngleSpherical
#
#      methods:
#             __call__: returns (jr,lz,jz)
#             actionsFreqs: returns (jr,lz,jz,Or,Op,Oz)
#             actionsFreqsAngles: returns (jr,lz,jz,Or,Op,Oz,ar,ap,az)
#
###############################################################################
import copy
import math as m
import numpy as nu
from scipy import integrate
from galpy.potential import epifreq, omegac, _dim
from galpy.potential.Potential import _evaluatePotentials
from galpy.potential.Potential import flatten as flatten_potential
from .actionAngle import *
from .actionAngleAxi import actionAngleAxi, potentialAxi
class actionAngleSpherical(actionAngle):
    """Action-angle formalism for spherical potentials"""
    def __init__(self,*args,**kwargs):
        """
        NAME:

           __init__

        PURPOSE:

           initialize an actionAngleSpherical object

        INPUT:

           pot= a Spherical potential

           ro= distance from vantage point to GC (kpc; can be Quantity)

           vo= circular velocity at ro (km/s; can be Quantity)

        OUTPUT:

           instance

        HISTORY:

           2013-12-28 - Written - Bovy (IAS)

        """
        actionAngle.__init__(self,
                             ro=kwargs.get('ro',None),vo=kwargs.get('vo',None))
        if not 'pot' in kwargs: #pragma: no cover
            raise IOError("Must specify pot= for actionAngleSpherical")
        self._pot= flatten_potential(kwargs['pot'])
        #Also store a 'planar' (2D) version of the potential
        if _dim(self._pot) == 2:
            self._2dpot= self._pot
        elif isinstance(self._pot,list):
            self._2dpot= [p.toPlanar() for p in self._pot]
        else:
            self._2dpot= self._pot.toPlanar()
        #The following for if we ever implement this code in C
        self._c= False
        ext_loaded= False
        if ext_loaded and (('c' in kwargs and kwargs['c'])
                           or not 'c' in kwargs):
            self._c= True #pragma: no cover
        else:
            self._c= False
        # Check the units
        self._check_consistent_units()
        return None

    def _evaluate(self,*args,**kwargs):
        """
        NAME:
           __call__ (_evaluate)
        PURPOSE:
           evaluate the actions (jr,lz,jz)
        INPUT:
           Either:
              a) R,vR,vT,z,vz[,phi]:
                 1) floats: phase-space value for single object (phi is optional) (each can be a Quantity)
                 2) numpy.ndarray: [N] phase-space values for N objects (each can be a Quantity)
              b) Orbit instance: initial condition used if that's it, orbit(t) if there is a time given as well as the second argument
           fixed_quad= (False) if True, use n=10 fixed_quad integration
           scipy.integrate.quadrature or .fixed_quad keywords
        OUTPUT:
           (jr,lz,jz)
        HISTORY:
           2013-12-28 - Written - Bovy (IAS)
        """
        fixed_quad= kwargs.pop('fixed_quad',False)
        if len(args) == 5: #R,vR.vT, z, vz
            R,vR,vT, z, vz= args
        elif len(args) == 6: #R,vR.vT, z, vz, phi
            R,vR,vT, z, vz, phi= args
        else:
            self._parse_eval_args(*args)
            R= self._eval_R
            vR= self._eval_vR
            vT= self._eval_vT
            z= self._eval_z
            vz= self._eval_vz
        if isinstance(R,float):
            R= nu.array([R])
            vR= nu.array([vR])
            vT= nu.array([vT])
            z= nu.array([z])
            vz= nu.array([vz])
        if self._c: #pragma: no cover
            pass
        else:
            Lz= R*vT
            Lx= -z*vT
            Ly= z*vR-R*vz
            L2= Lx*Lx+Ly*Ly+Lz*Lz
            E= _evaluatePotentials(self._pot,R,z)\
                +vR**2./2.+vT**2./2.+vz**2./2.
            L= nu.sqrt(L2)
            #Actions
            Jphi= Lz
            Jz= L-nu.fabs(Lz)
            #Jr requires some more work
            #Set up an actionAngleAxi object for EL and rap/rperi calculations
            axiR= nu.sqrt(R**2.+z**2.)
            axivT= L/axiR
            axivR= (R*vR+z*vz)/axiR
            Jr= []
            for ii in range(len(axiR)):
                axiaA= actionAngleAxi(axiR[ii],axivR[ii],axivT[ii],
                                      pot=self._2dpot)
                (rperi,rap)= axiaA.calcRapRperi()
                EL= axiaA.calcEL()
                E, L= EL
                Jr.append(self._calc_jr(rperi,rap,E,L,fixed_quad,**kwargs))
            return (nu.array(Jr),Jphi,Jz)

    def _actionsFreqs(self,*args,**kwargs):
        """
        NAME:
           actionsFreqs (_actionsFreqs)
        PURPOSE:
           evaluate the actions and frequencies (jr,lz,jz,Omegar,Omegaphi,Omegaz)
        INPUT:
           Either:
              a) R,vR,vT,z,vz[,phi]:
                 1) floats: phase-space value for single object (phi is optional) (each can be a Quantity)
                 2) numpy.ndarray: [N] phase-space values for N objects (each can be a Quantity)
              b) Orbit instance: initial condition used if that's it, orbit(t) if there is a time given as well as the second argument
           fixed_quad= (False) if True, use n=10 fixed_quad integration
           scipy.integrate.quadrature or .fixed_quad keywords
        OUTPUT:
            (jr,lz,jz,Omegar,Omegaphi,Omegaz)
        HISTORY:
           2013-12-28 - Written - Bovy (IAS)
        """
        fixed_quad= kwargs.pop('fixed_quad',False)
        if len(args) == 5: #R,vR.vT, z, vz
            R,vR,vT, z, vz= args
        elif len(args) == 6: #R,vR.vT, z, vz, phi
            R,vR,vT, z, vz, phi= args
        else:
            self._parse_eval_args(*args)
            R= self._eval_R
            vR= self._eval_vR
            vT= self._eval_vT
            z= self._eval_z
            vz= self._eval_vz
        if isinstance(R,float):
            R= nu.array([R])
            vR= nu.array([vR])
            vT= nu.array([vT])
            z= nu.array([z])
            vz= nu.array([vz])
        if self._c: #pragma: no cover
            pass
        else:
            Lz= R*vT
            Lx= -z*vT
            Ly= z*vR-R*vz
            L2= Lx*Lx+Ly*Ly+Lz*Lz
            E= _evaluatePotentials(self._pot,R,z)+vR**2./2.+vT**2./2.+vz**2./2.
            L= nu.sqrt(L2)
            #Actions
            Jphi= Lz
            Jz= L-nu.fabs(Lz)
            #Jr requires some more work
            #Set up an actionAngleAxi object for EL and rap/rperi calculations
            axiR= nu.sqrt(R**2.+z**2.)
            axivT= L/axiR
            axivR= (R*vR+z*vz)/axiR
            Jr= []
            Or= []
            Op= []
            for ii in range(len(axiR)):
                axiaA= actionAngleAxi(axiR[ii],axivR[ii],axivT[ii],
                                      pot=self._2dpot)
                (rperi,rap)= axiaA.calcRapRperi()
                EL= axiaA.calcEL()
                E, L= EL
                Jr.append(self._calc_jr(rperi,rap,E,L,fixed_quad,**kwargs))
                #Radial period
                if Jr[-1] < 10.**-9.: #Circular orbit
                    Or.append(epifreq(self._pot,axiR[ii],use_physical=False))
                    Op.append(omegac(self._pot,axiR[ii],use_physical=False))
                    continue
                Rmean= m.exp((m.log(rperi)+m.log(rap))/2.)
                Or.append(self._calc_or(Rmean,rperi,rap,E,L,fixed_quad,**kwargs))
                Op.append(self._calc_op(Or[-1],Rmean,rperi,rap,E,L,fixed_quad,**kwargs))
            Op= nu.array(Op)
            Oz= copy.copy(Op)
            Op[vT < 0.]*= -1.
            return (nu.array(Jr),Jphi,Jz,nu.array(Or),Op,Oz)
    
    def _actionsFreqsAngles(self,*args,**kwargs):
        """
        NAME:
           actionsFreqsAngles (_actionsFreqsAngles)
        PURPOSE:
           evaluate the actions, frequencies, and angles
           (jr,lz,jz,Omegar,Omegaphi,Omegaz,ar,ap,az)
        INPUT:
           Either:
              a) R,vR,vT,z,vz[,phi]:
                 1) floats: phase-space value for single object (phi is optional) (each can be a Quantity)
                 2) numpy.ndarray: [N] phase-space values for N objects (each can be a Quantity)
              b) Orbit instance: initial condition used if that's it, orbit(t) if there is a time given as well as the second argument
           fixed_quad= (False) if True, use n=10 fixed_quad integration
           scipy.integrate.quadrature or .fixed_quad keywords
        OUTPUT:
            (jr,lz,jz,Omegar,Omegaphi,Omegaz,ar,aphi,az)
        HISTORY:
           2013-12-29 - Written - Bovy (IAS)
        """
        fixed_quad= kwargs.pop('fixed_quad',False)
        if len(args) == 5: #R,vR.vT, z, vz pragma: no cover
            raise IOError("You need to provide phi when calculating angles")
        elif len(args) == 6: #R,vR.vT, z, vz, phi
            R,vR,vT, z, vz, phi= args
        else:
            self._parse_eval_args(*args)
            R= self._eval_R
            vR= self._eval_vR
            vT= self._eval_vT
            z= self._eval_z
            vz= self._eval_vz
            phi= self._eval_phi
        if isinstance(R,float):
            R= nu.array([R])
            vR= nu.array([vR])
            vT= nu.array([vT])
            z= nu.array([z])
            vz= nu.array([vz])
            phi= nu.array([phi])
        if self._c: #pragma: no cover
            pass
        else:
            Lz= R*vT
            Lx= -z*vT
            Ly= z*vR-R*vz
            L2= Lx*Lx+Ly*Ly+Lz*Lz
            E= _evaluatePotentials(self._pot,R,z)+vR**2./2.+vT**2./2.+vz**2./2.
            L= nu.sqrt(L2)
            #Actions
            Jphi= Lz
            Jz= L-nu.fabs(Lz)
            #Jr requires some more work
            #Set up an actionAngleAxi object for EL and rap/rperi calculations
            axiR= nu.sqrt(R**2.+z**2.)
            axivT= L/axiR #these are really spherical coords, called axi bc they go in actionAngleAxi
            axivR= (R*vR+z*vz)/axiR
            axivz= (z*vR-R*vz)/axiR
            Jr= []
            Or= []
            Op= []
            ar= []
            az= []
            #Calculate the longitude of the ascending node
            asc= self._calc_long_asc(z,R,axivz,phi,Lz,L)
            for ii in range(len(axiR)):
                axiaA= actionAngleAxi(axiR[ii],axivR[ii],axivT[ii],
                                      pot=self._2dpot)
                (rperi,rap)= axiaA.calcRapRperi()
                EL= axiaA.calcEL()
                E, L= EL
                Jr.append(self._calc_jr(rperi,rap,E,L,fixed_quad,**kwargs))
                #Radial period
                Rmean= m.exp((m.log(rperi)+m.log(rap))/2.)
                if Jr[-1] < 10.**-9.: #Circular orbit
                    Or.append(epifreq(self._pot,axiR[ii],use_physical=False))
                    Op.append(omegac(self._pot,axiR[ii],use_physical=False))
                else:
                    Or.append(self._calc_or(Rmean,rperi,rap,E,L,fixed_quad,**kwargs))
                    Op.append(self._calc_op(Or[-1],Rmean,rperi,rap,E,L,fixed_quad,**kwargs))
                #Angles
                ar.append(self._calc_angler(Or[-1],axiR[ii],Rmean,rperi,rap,
                                            E,L,axivR[ii],fixed_quad,**kwargs))
                az.append(self._calc_anglez(Or[-1],Op[-1],ar[-1],
                                            z[ii],axiR[ii],
                                            Rmean,rperi,rap,E,L,Lz[ii],
                                            axivR[ii],axivz[ii],
                                            fixed_quad,**kwargs))
            Op= nu.array(Op)
            Oz= copy.copy(Op)
            Op[vT < 0.]*= -1.
            ap= copy.copy(asc)
            ar= nu.array(ar)
            az= nu.array(az)
            ap[vT < 0.]-= az[vT < 0.]
            ap[vT >= 0.]+= az[vT >= 0.]
            ar= ar % (2.*nu.pi)
            ap= ap % (2.*nu.pi)
            az= az % (2.*nu.pi)
            return (nu.array(Jr),Jphi,Jz,nu.array(Or),Op,Oz,
                    ar,ap,az)
    
    def _EccZmaxRperiRap(self,*args,**kwargs):
        """
        NAME:
           EccZmaxRperiRap (_EccZmaxRperiRap)
        PURPOSE:
           evaluate the eccentricity, maximum height above the plane, peri- and apocenter for a spherical potential
        INPUT:
           Either:
              a) R,vR,vT,z,vz[,phi]:
                 1) floats: phase-space value for single object (phi is optional) (each can be a Quantity)
                 2) numpy.ndarray: [N] phase-space values for N objects (each can be a Quantity)
              b) Orbit instance: initial condition used if that's it, orbit(t) if there is a time given as well as the second argument
        OUTPUT:
           (e,zmax,rperi,rap)
        HISTORY:
           2017-12-22 - Written - Bovy (UofT)
        """
        if len(args) == 5: #R,vR.vT, z, vz
            R,vR,vT, z, vz= args
        elif len(args) == 6: #R,vR.vT, z, vz, phi
            R,vR,vT, z, vz, phi= args
        else:
            self._parse_eval_args(*args)
            R= self._eval_R
            vR= self._eval_vR
            vT= self._eval_vT
            z= self._eval_z
            vz= self._eval_vz
        if isinstance(R,float):
            R= nu.array([R])
            vR= nu.array([vR])
            vT= nu.array([vT])
            z= nu.array([z])
            vz= nu.array([vz])
        if self._c: #pragma: no cover
            pass
        else:
            Lz= R*vT
            Lx= -z*vT
            Ly= z*vR-R*vz
            L2= Lx*Lx+Ly*Ly+Lz*Lz
            L= nu.sqrt(L2)
            #Set up an actionAngleAxi object for EL and rap/rperi calculations
            axiR= nu.sqrt(R**2.+z**2.)
            axivT= L/axiR
            axivR= (R*vR+z*vz)/axiR
            rperi, rap= [], []
            for ii in range(len(axiR)):
                axiaA= actionAngleAxi(axiR[ii],axivR[ii],axivT[ii],
                                      pot=self._2dpot)
                trperi,trap= axiaA.calcRapRperi()
                rperi.append(trperi)
                rap.append(trap)
            rperi= nu.array(rperi)
            rap= nu.array(rap)
            return ((rap-rperi)/(rap+rperi),rap*nu.sqrt(1.-Lz**2./L2),
                    rperi,rap)

    def _calc_jr(self,rperi,rap,E,L,fixed_quad,**kwargs):
        if fixed_quad:
            return integrate.fixed_quad(_JrSphericalIntegrand,
                                        rperi,rap,
                                        args=(E,L,self._2dpot),
                                        n=10,
                                        **kwargs)[0]/nu.pi
        else:
            return (nu.array(integrate.quad(_JrSphericalIntegrand,
                                            rperi,rap,
                                            args=(E,L,self._2dpot),
                                            **kwargs)))[0]/nu.pi
    def _calc_or(self,Rmean,rperi,rap,E,L,fixed_quad,**kwargs):
        Tr= 0.
        if Rmean > rperi and not fixed_quad:
            Tr+= nu.array(integrate.quadrature(_TrSphericalIntegrandSmall,
                                               0.,m.sqrt(Rmean-rperi),
                                               args=(E,L,self._2dpot,
                                                     rperi),
                                               **kwargs))[0]
        elif Rmean > rperi and fixed_quad:
            Tr+= integrate.fixed_quad(_TrSphericalIntegrandSmall,
                                      0.,m.sqrt(Rmean-rperi),
                                      args=(E,L,self._2dpot,
                                            rperi),
                                      n=10,**kwargs)[0]
        if Rmean < rap and not fixed_quad:
            Tr+= nu.array(integrate.quadrature(_TrSphericalIntegrandLarge,
                                               0.,m.sqrt(rap-Rmean),
                                               args=(E,L,self._2dpot,
                                                     rap),
                                               **kwargs))[0]
        elif Rmean < rap and fixed_quad:
            Tr+= integrate.fixed_quad(_TrSphericalIntegrandLarge,
                                      0.,m.sqrt(rap-Rmean),
                                      args=(E,L,self._2dpot,
                                            rap),
                                      n=10,**kwargs)[0]
        Tr= 2.*Tr
        return 2.*nu.pi/Tr

    def _calc_op(self,Or,Rmean,rperi,rap,E,L,fixed_quad,**kwargs):
        #Azimuthal period
        I= 0.
        if Rmean > rperi and not fixed_quad:
            I+= nu.array(integrate.quadrature(_ISphericalIntegrandSmall,
                                              0.,m.sqrt(Rmean-rperi),
                                              args=(E,L,self._2dpot,
                                                    rperi),
                                              **kwargs))[0]
        elif Rmean > rperi and fixed_quad:
            I+= integrate.fixed_quad(_ISphericalIntegrandSmall,
                                     0.,m.sqrt(Rmean-rperi),
                                     args=(E,L,self._2dpot,rperi),
                                     n=10,**kwargs)[0]
        if Rmean < rap and not fixed_quad:
            I+= nu.array(integrate.quadrature(_ISphericalIntegrandLarge,
                                              0.,m.sqrt(rap-Rmean),
                                              args=(E,L,self._2dpot,
                                                    rap),
                                              **kwargs))[0]
        elif Rmean < rap and fixed_quad:
            I+= integrate.fixed_quad(_ISphericalIntegrandLarge,
                                     0.,m.sqrt(rap-Rmean),
                                     args=(E,L,self._2dpot,rap),
                                     n=10,**kwargs)[0]
        I*= 2*L
        return I*Or/2./nu.pi

    def _calc_long_asc(self,z,R,axivz,phi,Lz,L):
        i= nu.arccos(Lz/L)
        sinu= z/R/nu.tan(i)
        pindx= (sinu > 1.)*(sinu < (1.+10.**-7.))
        sinu[pindx]= 1.
        pindx= (sinu < -1.)*(sinu > (-1.-10.**-7.))
        sinu[pindx]= -1.           
        u= nu.arcsin(sinu)
        vzindx= axivz > 0.
        u[vzindx]= nu.pi-u[vzindx]
        return phi-u
    
    def _calc_angler(self,Or,r,Rmean,rperi,rap,E,L,vr,fixed_quad,**kwargs):
        if r < Rmean:
            if r > rperi and not fixed_quad:
                wr= Or*integrate.quadrature(_TrSphericalIntegrandSmall,
                                            0.,m.sqrt(r-rperi),
                                            args=(E,L,self._2dpot,rperi),
                                            **kwargs)[0]
            elif r > rperi and fixed_quad:
                wr= Or*integrate.fixed_quad(_TrSphericalIntegrandSmall,
                                            0.,m.sqrt(r-rperi),
                                            args=(E,L,self._2dpot,rperi),
                                            n=10,**kwargs)[0]
            else:
                wr= 0.
            if vr < 0.: wr= 2*m.pi-wr
        else:
            if r < rap and not fixed_quad:
                wr= Or*integrate.quadrature(_TrSphericalIntegrandLarge,
                                            0.,m.sqrt(rap-r),
                                            args=(E,L,self._2dpot,rap),
                                            **kwargs)[0]
            elif r < rap and fixed_quad:
                wr= Or*integrate.fixed_quad(_TrSphericalIntegrandLarge,
                                            0.,m.sqrt(rap-r),
                                            args=(E,L,self._2dpot,rap),
                                            n=10,**kwargs)[0]
            else:
                wr= m.pi
            if vr < 0.:
                wr= m.pi+wr
            else:
                wr= m.pi-wr
        return wr
        
    def _calc_anglez(self,Or,Op,ar,z,r,Rmean,rperi,rap,E,L,Lz,vr,axivz,
                     fixed_quad,**kwargs):
        #First calculate psi
        i= nu.arccos(Lz/L)
        sinpsi= z/r/nu.sin(i)
        if sinpsi > 1. and sinpsi < (1.+10.**-7.):
            sinpsi= 1.
        if sinpsi < -1. and sinpsi > (-1.-10.**-7.):
            sinpsi= -1.
        psi= nu.arcsin(sinpsi)
        if axivz > 0.: psi= nu.pi-psi
        psi= psi % (2.*nu.pi)
        #Calculate dSr/dL
        dpsi= Op/Or*2.*nu.pi #this is the full I integral
        if r < Rmean:
            if not fixed_quad:
                wz= L*integrate.quadrature(_ISphericalIntegrandSmall,
                                           0.,m.sqrt(r-rperi),
                                           args=(E,L,self._2dpot,
                                                 rperi),
                                           **kwargs)[0]
            elif fixed_quad:
                wz= L*integrate.fixed_quad(_ISphericalIntegrandSmall,
                                           0.,m.sqrt(r-rperi),
                                           args=(E,L,self._2dpot,
                                                 rperi),
                                           n=10,**kwargs)[0]
            if vr < 0.: wz= dpsi-wz
        else:
            if not fixed_quad:
                wz= L*integrate.quadrature(_ISphericalIntegrandLarge,
                                           0.,m.sqrt(rap-r),
                                           args=(E,L,self._2dpot,
                                                 rap),
                                           **kwargs)[0]
            elif fixed_quad:
                wz= L*integrate.fixed_quad(_ISphericalIntegrandLarge,
                                           0.,m.sqrt(rap-r),
                                           args=(E,L,self._2dpot,
                                                 rap),
                                           n=10,**kwargs)[0]
            if vr < 0.:
                wz= dpsi/2.+wz
            else:
                wz= dpsi/2.-wz
        #Add everything
        wz= -wz+psi+Op/Or*ar
        return wz

def _JrSphericalIntegrand(r,E,L,pot):
    """The J_r integrand"""
    return nu.sqrt(2.*(E-potentialAxi(r,pot))-L**2./r**2.)

def _TrSphericalIntegrandSmall(t,E,L,pot,rperi):
    r= rperi+t**2.#part of the transformation
    return 2.*t/_JrSphericalIntegrand(r,E,L,pot)

def _TrSphericalIntegrandLarge(t,E,L,pot,rap):
    r= rap-t**2.#part of the transformation
    return 2.*t/_JrSphericalIntegrand(r,E,L,pot)

def _ISphericalIntegrandSmall(t,E,L,pot,rperi):
    r= rperi+t**2.#part of the transformation
    return 2.*t/_JrSphericalIntegrand(r,E,L,pot)/r**2.

def _ISphericalIntegrandLarge(t,E,L,pot,rap):
    r= rap-t**2.#part of the transformation
    return 2.*t/_JrSphericalIntegrand(r,E,L,pot)/r**2.



###############################################################################
#   DehnenSmoothWrapperPotential.py: Wrapper to smoothly grow a potential
###############################################################################
from .WrapperPotential import parentWrapperPotential
from .Potential import _APY_LOADED
from galpy.util import bovy_conversion
if _APY_LOADED:
    from astropy import units
class DehnenSmoothWrapperPotential(parentWrapperPotential):
    """Potential wrapper class that implements the growth of a gravitational potential following `Dehnen (2000) <http://adsabs.harvard.edu/abs/2000AJ....119..800D>`__. The amplitude A applied to a potential wrapped by an instance of this class is changed as

    .. math::

        A(t) = amp\\,\\left(\\frac{3}{16}\\xi^5-\\frac{5}{8}\\xi^3+\\frac{15}{16}\\xi+\\frac{1}{2}\\right)

    where

    .. math::

        \\xi = \\begin{cases}
        -1 & t < t_\\mathrm{form}\\\\
        2\\left(\\frac{t-t_\\mathrm{form}}{t_\mathrm{steady}}\\right)-1\\,, &  t_\\mathrm{form} \\leq t \\leq t_\\mathrm{form}+t_\\mathrm{steady}\\\\
        1 & t > t_\\mathrm{form}+t_\\mathrm{steady}
        \\end{cases}

    if ``decay=True``, the amplitude decays rather than grows as decay = 1 - grow.
    """
    def __init__(self,amp=1.,pot=None,tform=-4.,tsteady=None,decay=False,
                 ro=None,vo=None):
        """
        NAME:

           __init__

        PURPOSE:

           initialize a DehnenSmoothWrapper Potential

        INPUT:

           amp - amplitude to be applied to the potential (default: 1.)

           pot - Potential instance or list thereof; the amplitude of this will be grown by this wrapper

           tform - start of growth (can be a Quantity)

           tsteady - time from tform at which the potential is fully grown (default: -tform/2, st the perturbation is fully grown at tform/2; can be a Quantity)

           decay= (False) if True, decay the amplitude instead of growing it (as 1-grow)

        OUTPUT:

           (none)

        HISTORY:

           2017-06-26 - Started - Bovy (UofT)

           2018-10-07 - Added 'decay' option - Bovy (UofT)

        """
        if _APY_LOADED and isinstance(tform,units.Quantity):
            tform= tform.to(units.Gyr).value\
                /bovy_conversion.time_in_Gyr(self._vo,self._ro)
        if _APY_LOADED and isinstance(tsteady,units.Quantity):
            tsteady= tsteady.to(units.Gyr).value\
                /bovy_conversion.time_in_Gyr(self._vo,self._ro)
        self._tform= tform
        if tsteady is None:
            self._tsteady= self._tform/2.
        else:
            self._tsteady= self._tform+tsteady
        self._grow= not decay
        self.hasC= True
        self.hasC_dxdv= True

    def _smooth(self,t):
        #Calculate relevant time
        if t < self._tform:
            smooth= 0.
        elif t < self._tsteady:
            deltat= t-self._tform
            xi= 2.*deltat/(self._tsteady-self._tform)-1.
            smooth= (3./16.*xi**5.-5./8*xi**3.+15./16.*xi+.5)
        else: #bar is fully on
            smooth= 1.
        return smooth if self._grow else 1.-smooth 

    def _wrap(self,attribute,*args,**kwargs):
        return self._smooth(kwargs.get('t',0.))\
                *self._wrap_pot_func(attribute)(self._pot,*args,**kwargs)



###############################################################################
#   actionAngle: a Python module to calculate  actions, angles, and frequencies
#
#      class: actionAngleVertical
#
#      methods:
#              Jz
#              anglez
#              Tz
#              calczmax
#              calcEz
###############################################################################
import math as m
import numpy as nu
from scipy import optimize, integrate
from .actionAngle import *
from galpy.potential.linearPotential import linearPotential, \
    evaluatelinearPotentials
class actionAngleVertical(actionAngle):
    """Action-angle formalism for vertical integral using the adiabatic approximation"""
    def __init__(self,*args,**kwargs):
        """
        NAME:
           __init__
        PURPOSE:
           initialize an actionAngleVertical object
        INPUT:
           Either:
              a) z,vz
              b) Orbit instance: initial condition used if that's it, orbit(t)
                 if there is a time given as well
              pot= potential or list of potentials (planarPotentials)
        OUTPUT:
        HISTORY:
           2012-06-01 - Written - Bovy (IAS)
        """
        self._parse_eval_args(*args,_noOrbUnitsCheck=True,**kwargs)
        self._z= self._eval_z
        self._vz= self._eval_vz
        if not 'pot' in kwargs: #pragma: no cover
            raise IOError("Must specify pot= for actionAngleVertical")
        self._verticalpot= kwargs['pot']
        return None
    
    def Jz(self,**kwargs):
        """
        NAME:
           Jz
        PURPOSE:
           Calculate the vertical action
        INPUT:
           +scipy.integrate.quad keywords
        OUTPUT:
           J_z(z,vz)/ro/vc + estimate of the error
        HISTORY:
           2012-06-01 - Written - Bovy (IAS)
        """
        if hasattr(self,'_Jz'):
            return self._Jz
        zmax= self.calczmax()
        if zmax == -9999.99: return nu.array([9999.99,nu.nan])
        Ez= calcEz(self._z,self._vz,self._verticalpot)
        self._Jz= 2.*integrate.quad(_JzIntegrand,0.,zmax,
                                    args=(Ez,self._verticalpot),
                                    **kwargs)[0]/nu.pi
        return self._Jz

    def Tz(self,**kwargs): #pragma: no cover
        """
        NAME:
           Tz
        PURPOSE:
           Calculate the vertical period
        INPUT:
           +scipy.integrate.quad keywords
        OUTPUT:
           T_z(z,vz)*vc/ro + estimate of the error
        HISTORY:
           2012-06-01 - Written - Bovy (IAS)
        """
        if hasattr(self,'_Tz'):
            return self._Tz
        zmax= self.calczmax()
        Ez= calcEz(self._z,self._vz,self._verticalpot)
        self._Tz= 4.*integrate.quad(_TzIntegrand,0.,zmax,
                                    args=(Ez,self._verticalpot),
                                    **kwargs)[0]
        return self._Tz

    def anglez(self,**kwargs): #pragma: no cover
        """
        NAME:
           anglez
        PURPOSE:
           Calculate the vertical angle
        INPUT:
           +scipy.integrate.quad keywords
        OUTPUT:
           angle_z(z,vz)*vc/ro + estimate of the error
        HISTORY:
           2012-06-01 - Written - Bovy (IAS)
        """
        if hasattr(self,'_anglez'):
            return self._anglez
        zmax= self.calczmax()
        Ez= calcEz(self._z,self._vz,self._verticalpot)
        Tz= self.Tz(**kwargs)
        self._anglez= 2.*nu.pi*(nu.array(integrate.quad(_TzIntegrand,0.,nu.fabs(self._z),
                                                        args=(Ez,self._verticalpot),
                                                        **kwargs)))/Tz[0]
        if self._z >= 0. and self._vz >= 0.:
            pass
        elif self._z >= 0. and self._vz < 0.:
            self._anglez[0]= nu.pi-self._anglez[0]
        elif self._z < 0. and self._vz <= 0.:
            self._anglez[0]= nu.pi+self._anglez[0]
        else:
            self._anglez[0]= 2.*nu.pi-self._anglez[0]
        return self._anglez

    def calczmax(self):
        """
        NAME:
           calczmax
        PURPOSE:
           calculate the maximum height
        INPUT:
        OUTPUT:
           zmax
        HISTORY:
           2012-06-01 - Written - Bovy (IAS)
        """
        if hasattr(self,'_zmax'): #pragma: no cover
            return self._zmax
        Ez= calcEz(self._z,self._vz,self._verticalpot)
        if self._vz == 0.: #We are exactly at the maximum height
            zmax= nu.fabs(self._z)
        else:
            zstart= self._z
            try:
                zend= _zmaxFindStart(self._z,Ez,self._verticalpot)
            except OverflowError: #pragma: no cover
                zmax= -9999.99
            else:
                zmax= optimize.brentq(_zmaxEq,zstart,zend,
                                      (Ez,self._verticalpot))
        self._zmax= zmax
        return self._zmax

def _zmaxEq(z,Ez,pot):
    """The vz=0 equation that needs to be solved to find zmax"""
    return Ez-potentialVertical(z,pot)

def calcEz(z,vz,pot):
    """
    NAME:
       calcEz
    PURPOSE:
       calculate the vertical energy
    INPUT:
       z - height (/ro)
       vz - vertical part of the velocity (/vc)
       pot - potential
    OUTPUT:
       Ez
    HISTORY:
       2012-06-01 - Written - Bovy (IAS)
    """                           
    return potentialVertical(z,pot)+vz**2./2.

def potentialVertical(z,pot):
    """
    NAME:
       potentialVertical
    PURPOSE:
       return the potential
    INPUT:
       z - height (/ro)
       pot - potential
    OUTPUT:
       Phi_z(z)
    HISTORY:
       2012-06-01 - Written - Bovy (IAS)
    """
    return evaluatelinearPotentials(pot,z,use_physical=False)

def _JzIntegrand(z,Ez,pot):
    """The J_z integrand"""
    return nu.sqrt(2.*(Ez-potentialVertical(z,pot)))

def _TzIntegrand(z,Ez,pot): #pragma: no cover
    """The T_z integrand"""
    return 1./_JzIntegrand(z,Ez,pot)

def _zmaxFindStart(z,Ez,pot):
    """
    NAME:
       _zmaxFindStart
    PURPOSE:
       Find adequate end point to solve for zmax
    INPUT:
       z - height
       Ez - vertical energy
       pot - potential
    OUTPUT:
       zend
    HISTORY:
       2012-06-01 - Written - Bovy (IAS)
    """
    if z == 0.: ztry= 0.00001
    else: ztry= 2.*nu.fabs(z)
    while (Ez-potentialVertical(ztry,pot)) > 0.:
        ztry*= 2.
        if ztry > 100.: #pragma: no cover
            raise OverflowError
    return ztry




###############################################################################
#   BurkertPotential.py: Potential with a Burkert density
###############################################################################
import numpy
from .Potential import Potential, _APY_LOADED
if _APY_LOADED:
    from astropy import units
class BurkertPotential(Potential):
    """BurkertPotential.py: Potential with a Burkert density

    .. math::

        \\rho(r) = \\frac{\\mathrm{amp}}{(1+r/a)\\,(1+[r/a]^2)}

    """
    def __init__(self,amp=1.,a=2.,normalize=False,
                 ro=None,vo=None):
        """
        NAME:

           __init__

        PURPOSE:

           initialize a Burkert-density potential

        INPUT:

           amp - amplitude to be applied to the potential (default: 1); can be a Quantity with units of mass density or Gxmass density

           a = scale radius (can be Quantity)

           normalize - if True, normalize such that vc(1.,0.)=1., or, if 
                       given as a number, such that the force is this fraction 
                       of the force necessary to make vc(1.,0.)=1.

           ro=, vo= distance and velocity scales for translation into internal units (default from configuration file)

        OUTPUT:

           (none)

        HISTORY:

           2013-04-10 - Written - Bovy (IAS)

        """
        Potential.__init__(self,amp=amp,ro=ro,vo=vo,amp_units='density')
        if _APY_LOADED and isinstance(a,units.Quantity):
            a= a.to(units.kpc).value/self._ro
        self.a=a
        self._scale= self.a
        if normalize or \
                (isinstance(normalize,(int,float)) \
                     and not isinstance(normalize,bool)): #pragma: no cover 
            self.normalize(normalize)
        self.hasC= True
        self.hasC_dxdv= True
        return None

    def _evaluate(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _evaluate
        PURPOSE:
           evaluate the potential at R,z
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           Phi(R,z)
        HISTORY:
           2013-04-10 - Started - Bovy (IAS)
        """
        x= numpy.sqrt(R**2.+z**2.)/self.a
        return -self.a**2.*numpy.pi/x*(-numpy.pi+2.*(1.+x)*numpy.arctan(1/x)+2.*(1.+x)*numpy.log(1.+x)+(1.-x)*numpy.log(1.+x**2.))

    def _Rforce(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _Rforce
        PURPOSE:
           evaluate the radial force for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           the radial force
        HISTORY:
           2013-04-10 - Written - Bovy (IAS)
        """
        r= numpy.sqrt(R**2.+z**2.)
        x= r/self.a
        return self.a*numpy.pi/x**2.*(numpy.pi-2.*numpy.arctan(1./x)-2.*numpy.log(1.+x)-numpy.log(1.+x**2.))*R/r

    def _zforce(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _zforce
        PURPOSE:
           evaluate the vertical force for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           the vertical force
        HISTORY:
           2013-04-10 - Written - Bovy (IAS)
        """
        r= numpy.sqrt(R**2.+z**2.)
        x= r/self.a
        return self.a*numpy.pi/x**2.*(numpy.pi-2.*numpy.arctan(1./x)-2.*numpy.log(1.+x)-numpy.log(1.+x**2.))*z/r

    def _R2deriv(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _Rderiv
        PURPOSE:
           evaluate the second radial derivative for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           the second radial derivative
        HISTORY:
           2013-04-10 - Written - Bovy (IAS)
        """
        r= numpy.sqrt(R**2.+z**2.)
        x= r/self.a
        return -numpy.pi/x**3./r**2.*(-4.*R**2.*r**3./(self.a**2.+r**2.)/(self.a+r)+(z**2.-2.*R**2.)*(numpy.pi-2.*numpy.arctan(1./x)-2.*numpy.log(1.+x)-numpy.log(1.+x**2.)))

    def _z2deriv(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _z2deriv
        PURPOSE:
           evaluate the second vertical derivative for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t- time
        OUTPUT:
           the second vertical derivative
        HISTORY:
           2012-07-26 - Written - Bovy (IAS@MPIA)
        """
        return self._R2deriv(z,R) #Spherical potential

    def _dens(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _dens
        PURPOSE:
           evaluate the density for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           the density
        HISTORY:
           2013-01-09 - Written - Bovy (IAS)
        """
        r= numpy.sqrt(R**2.+z**2.)
        x= r/self.a
        return 1./(1.+x)/(1.+x**2.)

    def _surfdens(self,R,z,phi=0.,t=0.):
        """
        NAME:
           _surfdens
        PURPOSE:
           evaluate the surface density for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           z - vertical height
           phi - azimuth
           t - time
        OUTPUT:
           the surface density
        HISTORY:
           2018-08-19 - Written - Bovy (UofT)
        """
        r= numpy.sqrt(R**2.+z**2.)
        x= r/self.a
        Rpa= numpy.sqrt(R**2.+self.a**2.)
        Rma= numpy.sqrt(R**2.-self.a**2.+0j)
        if Rma == 0:
            za= z/self.a
            return self.a**2./2.*((2.-2.*numpy.sqrt(za**2.+1)
                                   +numpy.sqrt(2.)*za\
                                       *numpy.arctan(za/numpy.sqrt(2.)))/z
                                  +numpy.sqrt(2*za**2.+2.)\
                                   *numpy.arctanh(za/numpy.sqrt(2.*(za**2.+1)))
                                  /numpy.sqrt(self.a**2.+z**2.))
        else:
            return self.a**2.*(numpy.arctan(z/x/Rma)/Rma
                               +numpy.arctanh(z/x/Rpa)/Rpa
                               -numpy.arctan(z/Rma)/Rma
                               +numpy.arctan(z/Rpa)/Rpa).real



###############################################################################
#   actionAngle: a Python module to calculate  actions, angles, and frequencies
#
#      class: actionAngleAdiabatic
#
#             wrapper around actionAngleAxi (adiabatic approximation) to do
#             this for any (x,v)
#
#      methods:
#             __call__: returns (jr,lz,jz)
#
###############################################################################
import copy
import warnings
import math as m
import numpy as nu
from galpy.util import galpyWarning
from galpy.potential import planarPotential, MWPotential
from galpy.potential.Potential import flatten as flatten_potential
from ..potential import toPlanarPotential, toVerticalPotential
from .actionAngleAxi import actionAngleAxi
from .actionAngle import actionAngle
from . import actionAngleAdiabatic_c
from .actionAngleAdiabatic_c import _ext_loaded as ext_loaded
from galpy.potential.Potential import _check_c
class actionAngleAdiabatic(actionAngle):
    """Action-angle formalism for axisymmetric potentials using the adiabatic approximation"""
    def __init__(self,*args,**kwargs):
        """
        NAME:

           __init__

        PURPOSE:

           initialize an actionAngleAdiabatic object

        INPUT:

           pot= potential or list of potentials (planarPotentials)

           gamma= (default=1.) replace Lz by Lz+gamma Jz in effective potential

           ro= distance from vantage point to GC (kpc; can be Quantity)

           vo= circular velocity at ro (km/s; can be Quantity)

        OUTPUT:
        
           instance

        HISTORY:

            2012-07-26 - Written - Bovy (IAS@MPIA)

        """
        actionAngle.__init__(self,
                             ro=kwargs.get('ro',None),vo=kwargs.get('vo',None))
        if not 'pot' in kwargs: #pragma: no cover
            raise IOError("Must specify pot= for actionAngleAxi")
        self._pot= flatten_potential(kwargs['pot'])
        if self._pot == MWPotential:
            warnings.warn("Use of MWPotential as a Milky-Way-like potential is deprecated; galpy.potential.MWPotential2014, a potential fit to a large variety of dynamical constraints (see Bovy 2015), is the preferred Milky-Way-like potential in galpy",
                          galpyWarning)
        if ext_loaded and 'c' in kwargs and kwargs['c']:
            self._c= _check_c(self._pot)
            if 'c' in kwargs and kwargs['c'] and not self._c:
                warnings.warn("C module not used because potential does not have a C implementation",galpyWarning) #pragma: no cover
        else:
            self._c= False
        self._gamma= kwargs.get('gamma',1.)
        # Check the units
        self._check_consistent_units()
        return None
    
    def _evaluate(self,*args,**kwargs):
        """
        NAME:
           __call__ (_evaluate)
        PURPOSE:
           evaluate the actions (jr,lz,jz)
        INPUT:
           Either:
              a) R,vR,vT,z,vz[,phi]:
                 1) floats: phase-space value for single object (phi is optional) (each can be a Quantity)
                 2) numpy.ndarray: [N] phase-space values for N objects (each can be a Quantity)
              b) Orbit instance: initial condition used if that's it, orbit(t) if there is a time given as well as the second argument 
           c= (object-wide default, bool) True/False to override the object-wide setting for whether or not to use the C implementation
           scipy.integrate.quadrature keywords
           _justjr, _justjz= if True, only calculate the radial or vertical action (internal use)
        OUTPUT:
           (jr,lz,jz)
        HISTORY:
           2012-07-26 - Written - Bovy (IAS@MPIA)
        """
        if len(args) == 5: #R,vR.vT, z, vz
            R,vR,vT, z, vz= args
        elif len(args) == 6: #R,vR.vT, z, vz, phi
            R,vR,vT, z, vz, phi= args
        else:
            self._parse_eval_args(*args)
            R= self._eval_R
            vR= self._eval_vR
            vT= self._eval_vT
            z= self._eval_z
            vz= self._eval_vz
        if isinstance(R,float):
            R= nu.array([R])
            vR= nu.array([vR])
            vT= nu.array([vT])
            z= nu.array([z])
            vz= nu.array([vz])
        if ((self._c and not ('c' in kwargs and not kwargs['c']))\
                or (ext_loaded and (('c' in kwargs and kwargs['c'])))) \
                and _check_c(self._pot):
            Lz= R*vT
            jr, jz, err= actionAngleAdiabatic_c.actionAngleAdiabatic_c(\
                self._pot,self._gamma,R,vR,vT,z,vz)
            if err == 0:
                return (jr,Lz,jz)
            else: #pragma: no cover
                raise RuntimeError("C-code for calculation actions failed; try with c=False")
        else:
            if 'c' in kwargs and kwargs['c'] and not self._c:
                warnings.warn("C module not used because potential does not have a C implementation",galpyWarning) #pragma: no cover
            kwargs.pop('c',None)
            if len(R) > 1:
                ojr= nu.zeros((len(R)))
                olz= nu.zeros((len(R)))
                ojz= nu.zeros((len(R)))
                for ii in range(len(R)):
                    targs= (R[ii],vR[ii],vT[ii],z[ii],vz[ii])
                    tjr,tlz,tjz= self(*targs,**copy.copy(kwargs))
                    ojr[ii]= tjr
                    ojz[ii]= tjz
                    olz[ii]= tlz
                return (ojr,olz,ojz)
            else:
                #Set up the actionAngleAxi object
                thispot= toPlanarPotential(self._pot)
                thisverticalpot= toVerticalPotential(self._pot,R[0])
                aAAxi= actionAngleAxi(R[0],vR[0],vT[0],z[0],vz[0],
                                      pot=thispot,
                                       verticalPot=thisverticalpot,
                                       gamma=self._gamma)
                if kwargs.get('_justjr',False):
                    kwargs.pop('_justjr')
                    return (aAAxi.JR(**kwargs),nu.nan,nu.nan)
                elif kwargs.get('_justjz',False):
                    kwargs.pop('_justjz')
                    return (nu.atleast_1d(nu.nan),
                            nu.atleast_1d(nu.nan),
                            nu.atleast_1d(aAAxi.Jz(**kwargs)))
                else:
                    return (nu.atleast_1d(aAAxi.JR(**kwargs)),
                            nu.atleast_1d(aAAxi._R*aAAxi._vT),
                            nu.atleast_1d(aAAxi.Jz(**kwargs)))

    def _EccZmaxRperiRap(self,*args,**kwargs):
        """
        NAME:
           EccZmaxRperiRap (_EccZmaxRperiRap)
        PURPOSE:
           evaluate the eccentricity, maximum height above the plane, peri- and apocenter in the adiabatic approximation
        INPUT:
           Either:
              a) R,vR,vT,z,vz[,phi]:
                 1) floats: phase-space value for single object (phi is optional) (each can be a Quantity)
                 2) numpy.ndarray: [N] phase-space values for N objects (each can be a Quantity)
              b) Orbit instance: initial condition used if that's it, orbit(t) if there is a time given as well as the second argument 
           c= (object-wide default, bool) True/False to override the object-wide setting for whether or not to use the C implementation
        OUTPUT:
           (e,zmax,rperi,rap)
        HISTORY:
           2017-12-21 - Written - Bovy (UofT)
        """
        if len(args) == 5: #R,vR.vT, z, vz
            R,vR,vT, z, vz= args
        elif len(args) == 6: #R,vR.vT, z, vz, phi
            R,vR,vT, z, vz, phi= args
        else:
            self._parse_eval_args(*args)
            R= self._eval_R
            vR= self._eval_vR
            vT= self._eval_vT
            z= self._eval_z
            vz= self._eval_vz
        if isinstance(R,float):
            R= nu.array([R])
            vR= nu.array([vR])
            vT= nu.array([vT])
            z= nu.array([z])
            vz= nu.array([vz])
        if ((self._c and not ('c' in kwargs and not kwargs['c']))\
                or (ext_loaded and (('c' in kwargs and kwargs['c'])))) \
                and _check_c(self._pot):
            rperi,Rap,zmax, err= actionAngleAdiabatic_c.actionAngleRperiRapZmaxAdiabatic_c(\
                self._pot,self._gamma,R,vR,vT,z,vz)
            if err == 0:
                rap= nu.sqrt(Rap**2.+zmax**2.)
                ecc= (rap-rperi)/(rap+rperi)
                return (ecc,zmax,rperi,rap)
            else: #pragma: no cover
                raise RuntimeError("C-code for calculation actions failed; try with c=False")
        else:
            if 'c' in kwargs and kwargs['c'] and not self._c:
                warnings.warn("C module not used because potential does not have a C implementation",galpyWarning) #pragma: no cover
            kwargs.pop('c',None)
            if len(R) > 1:
                oecc= nu.zeros((len(R)))
                orperi= nu.zeros((len(R)))
                orap= nu.zeros((len(R)))
                ozmax= nu.zeros((len(R)))
                for ii in range(len(R)):
                    targs= (R[ii],vR[ii],vT[ii],z[ii],vz[ii])
                    tecc, tzmax, trperi,trap= self._EccZmaxRperiRap(\
                        *targs,**copy.copy(kwargs))
                    oecc[ii]= tecc
                    ozmax[ii]= tzmax
                    orperi[ii]= trperi
                    orap[ii]= trap
                return (oecc,ozmax,orperi,orap)
            else:
                #Set up the actionAngleAxi object
                thispot= toPlanarPotential(self._pot)
                thisverticalpot= toVerticalPotential(self._pot,R[0])
                aAAxi= actionAngleAxi(R[0],vR[0],vT[0],z[0],vz[0],
                                      pot=thispot,
                                       verticalPot=thisverticalpot,
                                       gamma=self._gamma)
                rperi,Rap= aAAxi.calcRapRperi(**kwargs)
                zmax= aAAxi.calczmax(**kwargs)
                rap= nu.sqrt(Rap**2.+zmax**2.)
                return (nu.atleast_1d((rap-rperi)/(rap+rperi)),
                        nu.atleast_1d(zmax),nu.atleast_1d(rperi),
                        nu.atleast_1d(rap))

    def calcRapRperi(self,*args,**kwargs):
        """
        NAME:
           calcRapRperi
        PURPOSE:
           calculate the apocenter and pericenter radii
        INPUT:
           Either:
              a) R,vR,vT,z,vz
              b) Orbit instance: initial condition used if that's it, orbit(t)
                 if there is a time given as well
        OUTPUT:
           (rperi,rap)
        HISTORY:
           2013-11-27 - Written - Bovy (IAS)
        """
        #Set up the actionAngleAxi object
        thispot= toPlanarPotential(self._pot)
        aAAxi= actionAngleAxi(*args,pot=thispot,amma=self._gamma)
        return aAAxi.calcRapRperi(**kwargs)

    def calczmax(self,*args,**kwargs): #pragma: no cover
        """
        NAME:
           calczmax
        PURPOSE:
           calculate the maximum height
        INPUT:
           Either:
              a) R,vR,vT,z,vz
              b) Orbit instance: initial condition used if that's it, orbit(t)
                 if there is a time given as well
        OUTPUT:
           zmax
        HISTORY:
           2012-06-01 - Written - Bovy (IAS)
        """
        warnings.warn("actionAngleAdiabatic.calczmax function will soon be deprecated; please contact galpy's maintainer if you require this function")
        #Set up the actionAngleAxi object
        self._parse_eval_args(*args)
        thispot= toPlanarPotential(self._pot)
        thisverticalpot= toVerticalPotential(self._pot,self._eval_R)
        aAAxi= actionAngleAxi(*args,pot=thispot,
                               verticalPot=thisverticalpot,
                               gamma=self._gamma)
        return aAAxi.calczmax(**kwargs)



###############################################################################
#   HenonHeilesPotential: the Henon-Heiles (1964) potential
###############################################################################
import numpy
from .planarPotential import planarPotential
class HenonHeilesPotential(planarPotential):
    """Class that implements a the `Henon & Heiles (1964) <http://adsabs.harvard.edu/abs/1964AJ.....69...73H>`__ potential
    
    .. math::

        \\Phi(R,\\phi) = \\frac{\\mathrm{amp}}{2}\\,\\left[R^2 + \\frac{2\\,R^3}{3}\\,\\sin\\left(3\,\phi\\right)\\right]

    """
    def __init__(self,amp=1.,ro=None,vo=None):
        """
        NAME:

           __init__

        PURPOSE:

           initialize a Henon-Heiles potential

        INPUT:

           amp - amplitude to be applied to the potential (default: 1.)

        OUTPUT:

           (none)

        HISTORY:

           2017-10-16 - Written - Bovy (UofT)

        """
        planarPotential.__init__(self,amp=amp,ro=ro,vo=vo)
        self.hasC= True
        self.hasC_dxdv= True

    def _evaluate(self,R,phi=0.,t=0.):
        """
        NAME:
           _evaluate
        PURPOSE:
           evaluate the potential at R,phi,t
        INPUT:
           R - Galactocentric cylindrical radius
           phi - azimuth
           t - time
        OUTPUT:
           Phi(R,phi,t)
        HISTORY:
           2017-10-16 - Written - Bovy (UofT)
        """
        return 0.5*R*R*(1.+2./3.*R*numpy.sin(3.*phi))

    def _Rforce(self,R,phi=0.,t=0.):
        """
        NAME:
           _Rforce
        PURPOSE:
           evaluate the radial force for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           phi - azimuth
           t - time
        OUTPUT:
           the radial force
        HISTORY:
           2017-10-16 - Written - Bovy (UofT)
        """
        return -R*(1.+R*numpy.sin(3.*phi))
       
    def _phiforce(self,R,phi=0.,t=0.):
        """
        NAME:
           _phiforce
        PURPOSE:
           evaluate the azimuthal force for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           phi - azimuth
           t - time
        OUTPUT:
           the azimuthal force
        HISTORY:
           2017-10-16 - Written - Bovy (UofT)
        """
        return -R**3.*numpy.cos(3.*phi)

    def _R2deriv(self,R,phi=0.,t=0.):
        """
        NAME:
           _R2deriv
        PURPOSE:
           evaluate the second radial derivative for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           phi - azimuth
           t - time
        OUTPUT:
           the second radial derivative
        HISTORY:
           2017-10-16 - Written - Bovy (UofT)
        """
        return 1.+2.*R*numpy.sin(3.*phi)
       
    def _phi2deriv(self,R,phi=0.,t=0.):
        """
        NAME:
           _phi2deriv
        PURPOSE:
           evaluate the second azimuthal derivative for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           phi - azimuth
           t - time
        OUTPUT:
           the second azimuthal derivative
        HISTORY:
           2017-10-16 - Written - Bovy (UofT)
        """
        return -3.*R**3.*numpy.sin(3.*phi)

    def _Rphideriv(self,R,phi=0.,t=0.):
        """
        NAME:
           _Rphideriv
        PURPOSE:
           evaluate the mixed radial, azimuthal derivative for this potential
        INPUT:
           R - Galactocentric cylindrical radius
           phi - azimuth
           t - time
        OUTPUT:
           the mixed radial, azimuthal derivative
        HISTORY:
           2017-10-16 - Written - Bovy (UofT)
        """
        return 3.*R**2.*numpy.cos(3.*phi)




try:
    import pynbody
    _PYNBODYENABLED= True
except ImportError:
    _PYNBODYENABLED= False
class GadgetSnapshot(object):
    """Snapshot coming out of gadget"""
    def __init__(self,*args,**kwargs):
        """
        NAME:
           __init__
        PURPOSE:
           initialize a Gadget snapshot object
        INPUT:
           Initialize using:
              1) filename
        OUTPUT:
        HISTORY:
           2011-08-15 - Started - Bovy
        """
        if not _PYNBODYENABLED:
            raise ImportError("pynbody could not be loaded to read the gadget snapshot")
        filename= args[0]
        self._sim = pynbody.load(filename)
        self._sim.physical_units()
        self._star= self._sim.star
        return None

    def __getattr__(self,i):
        if i == 'x':
            return self.pos[0]
        elif i =='pos':
            return self._star['pos']

    def plot(self,*args,**kwargs):
        """
        NAME:
           plot
        PURPOSE:
           plot the snapshot
        INPUT:
        
        OUTPUT:
        HISTORY:
           2011-08-15 - Started - Bovy (NYU)
        """
        labeldict= {'t':r'$t$','R':r'$R$','vR':r'$v_R$','vT':r'$v_T$',
                    'z':r'$z$','vz':r'$v_z$','phi':r'$\phi$',
                    'x':r'$x$','y':r'$y$','vx':r'$v_x$','vy':r'$v_y$'}
        #Defaults
        if not kwargs.has_key('d1') and not kwargs.has_key('d2'):
            if len(self.orbits[0].vxvv) == 3:
                d1= 'R'
                d2= 'vR'
            elif len(self.orbits[0].vxvv) == 4:
                d1= 'x'
                d2= 'y'
            elif len(self.orbits[0].vxvv) == 2:
                d1= 'x'
                d2= 'vx'
            elif len(self.orbits[0].vxvv) == 5 \
                    or len(self.orbits[0].vxvv) == 6:
                d1= 'R'
                d2= 'z'
        elif not kwargs.has_key('d1'):
            d2= kwargs['d2']
            kwargs.pop('d2')
            d1= 't'
        elif not kwargs.has_key('d2'):
            d1= kwargs['d1']
            kwargs.pop('d1')
            d2= 't'
        else:
            d1= kwargs['d1']
            kwargs.pop('d1')
            d2= kwargs['d2']
            kwargs.pop('d2')




# jeans.py: utilities related to the Jeans equations
import numpy
from scipy import integrate
from galpy.potential.Potential import evaluateDensities, \
    evaluaterforces, evaluateSurfaceDensities
from galpy.potential.Potential import flatten as flatten_pot
from galpy.util.bovy_conversion import physical_conversion, \
    potential_physical_input
_INVSQRTTWO= 1./numpy.sqrt(2.)
@potential_physical_input
@physical_conversion('velocity',pop=True)
def sigmar(Pot,r,dens=None,beta=0.):
    """
    NAME:

       sigmar

    PURPOSE:

       Compute the radial velocity dispersion using the spherical Jeans equation

    INPUT:

       Pot - potential or list of potentials (evaluated at R=r/sqrt(2),z=r/sqrt(2), sphericity not checked)

       r - Galactocentric radius (can be Quantity)

       dens= (None) tracer density profile (function of r); if None, the density is assumed to be that corresponding to the potential

       beta= (0.) anisotropy; can be a constant or a function of r
       
    OUTPUT:

       sigma_r(r)

    HISTORY:

       2018-07-05 - Written - Bovy (UofT)

    """
    Pot= flatten_pot(Pot)
    if dens is None:
        dens= lambda r: evaluateDensities(Pot,r*_INVSQRTTWO,r*_INVSQRTTWO,
                                          use_physical=False)
    if callable(beta):
        intFactor= lambda x: numpy.exp(2.*integrate.quad(lambda y: beta(y)/y,
                                       1.,x)[0])
    else: # assume to be number
        intFactor= lambda x: x**(2.*beta)
    return numpy.sqrt(integrate.quad(lambda x: -intFactor(x)*dens(x)
                                     *evaluaterforces(Pot,
                                                      x*_INVSQRTTWO,
                                                      x*_INVSQRTTWO,
                                                      use_physical=False),
                                     r,numpy.inf)[0]/
                      dens(r)/intFactor(r))

@potential_physical_input
@physical_conversion('velocity',pop=True)
def sigmalos(Pot,R,dens=None,surfdens=None,beta=0.,sigma_r=None):
    """
    NAME:

       sigmalos

    PURPOSE:

       Compute the line-of-sight velocity dispersion using the spherical Jeans equation

    INPUT:

       Pot - potential or list of potentials (evaluated at R=r/sqrt(2),z=r/sqrt(2), sphericity not checked)

       R - Galactocentric projected radius (can be Quantity)

       dens= (None) tracer density profile (function of r); if None, the density is assumed to be that corresponding to the potential

       surfdens= (None) tracer surface density profile (value at R or function of R); if None, the surface density is assumed to be that corresponding to the density

       beta= (0.) anisotropy; can be a constant or a function of r

       sigma_r= (None) if given, the solution of the spherical Jeans equation sigma_r(r) (used instead of solving the Jeans equation as part of this routine)
       
    OUTPUT:

       sigma_los(R)

    HISTORY:

       2018-08-27 - Written - Bovy (UofT)

    """
    Pot= flatten_pot(Pot)
    if dens is None:
        densPot= True
        dens= lambda r: evaluateDensities(Pot,r*_INVSQRTTWO,r*_INVSQRTTWO,
                                          use_physical=False)
    else:
        densPot= False
    if callable(surfdens):
        called_surfdens= surfdens(R)
    elif surfdens is None:
        if densPot:
            called_surfdens= evaluateSurfaceDensities(Pot,R,numpy.inf,
                                                    use_physical=False)
        if not densPot or numpy.isnan(called_surfdens):
            called_surfdens=\
                    2.*integrate.quad(lambda x: dens(numpy.sqrt(R**2.+x**2.)),
                                      0.,numpy.inf)[0]
    else:
        called_surfdens= surfdens
    if callable(beta):
        call_beta= beta
    else:
        call_beta= lambda x: beta
    if sigma_r is None:
        call_sigma_r= lambda r: sigmar(Pot,r,dens=dens,beta=beta)
    elif not callable(sigma_r):
        call_sigma_r= lambda x: sigma_r
    else:
        call_sigma_r= sigma_r
    return numpy.sqrt(2.*integrate.quad(\
            lambda x: (1.-call_beta(x)*R**2./x**2.)*x*dens(x)\
                *call_sigma_r(x)**2./numpy.sqrt(x**2.-R**2.),R,numpy.inf)[0]\
                          /called_surfdens)
